import 'dart:convert';
import 'dart:html' as html;
import 'dart:typed_data';

import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter_markdown/flutter_markdown.dart';
import 'package:google_fonts/google_fonts.dart';
import 'package:intl/intl.dart';
import 'package:remove_emoji/remove_emoji.dart';
import 'package:markdown/markdown.dart' as md;
import 'dart:ui' as ui;
import 'package:url_launcher/url_launcher.dart';
import 'package:dotted_border/dotted_border.dart';
import 'package:drop_zone/drop_zone.dart';

abstract class Parser {
  Widget parse(Map block, {Options options, BuildContext context});
}

class Options {
  Options({
    this.disabled = false,
    this.alignment,
    this.margin = const EdgeInsets.all(0),
    this.padding = const EdgeInsets.all(0),
    this.textAlign = TextAlign.center,
    this.textStyle =
        const TextStyle(fontFamily: "Raleway", fontSize: 14, fontWeight: FontWeight.w400, color: Colors.black),
    this.decoration = const BoxDecoration(),
  });

  bool disabled;
  AlignmentGeometry alignment;
  EdgeInsets margin;
  EdgeInsets padding;
  TextAlign textAlign;
  TextStyle textStyle;
  Decoration decoration;
}

class PlainTextWidget extends StatefulWidget {
  final Map block;
  final Options options;

  PlainTextWidget(this.block, {this.options});

  @override
  createState() => PlainTextWidgetState();
}

class PlainTextWidgetState extends State<PlainTextWidget> {
  @override
  Widget build(BuildContext context) {
    if (widget.block == null) return Container();

    var hasEmoji = widget.block["emoji"];
    var innerType = widget.block["type"];
    var innerText = widget.block["text"];
    if (hasEmoji != null && hasEmoji == false) {
      innerText = RemoveEmoji().removemoji(innerText);
    }

    Widget innerWidget = Container();
    if (innerType == "plain_text") {
      innerWidget = Text(innerText, style: widget.options.textStyle, textAlign: widget.options.textAlign);
    } else if (innerType == "mrkdwn") {
      innerWidget = MarkdownBody(
        data: innerText,
        extensionSet: md.ExtensionSet(
          md.ExtensionSet.gitHubFlavored.blockSyntaxes,
          [md.EmojiSyntax(), ...md.ExtensionSet.gitHubFlavored.inlineSyntaxes],
        ),
        styleSheet:
            MarkdownStyleSheet.fromTheme(Theme.of(context)).copyWith(p: widget.options.textStyle.copyWith(height: 1.4)),
        onTapLink: (text, href, title) {
          launch(href);
        },
      );
    }

    return Container(
      alignment: widget.options.alignment,
      padding: widget.options.padding,
      decoration: widget.options.decoration,
      child: innerWidget,
    );
  }
}

class PlainTextParser implements Parser {
  @override
  Widget parse(Map block, {Options options, BuildContext context}) {
    return PlainTextWidget(block, options: options ?? Options());
  }
}

class ButtonParser implements Parser {
  @override
  Widget parse(Map block, {Options options, BuildContext context}) {
    //var action = block["action"] != null ? block["action"] : {};
    var text = block["text"];

    Widget inner = PlainTextParser().parse(text,
        context: context,
        options: Options(
            decoration: BoxDecoration(
                border: Border.all(width: 0.5, color: Colors.grey.shade600),
                borderRadius: BorderRadius.all(Radius.circular(4))),
            padding: EdgeInsets.fromLTRB(8, 6, 8, 6),
            textStyle:
                GoogleFonts.getFont('Open Sans', fontSize: 14, fontWeight: FontWeight.w600, color: Colors.black)));

    return Container(
        margin: options.margin,
        child: TextButton(
          onPressed: () {},
          child: inner,
          style: ButtonStyle(overlayColor: MaterialStateProperty.all(Colors.grey.shade100)),
        ));
  }
}

class SpacerParser implements Parser {
  @override
  Widget parse(Map block, {Options options, BuildContext context}) {
    return Spacer();
  }
}

class SectionParser implements Parser {
  const SectionParser();

  @override
  Widget parse(Map block, {Options options, BuildContext context}) {
    var tb = block["text"];
    if (tb != null) {
      return PlainTextParser().parse(tb,
          context: context,
          options: Options(
              textAlign: TextAlign.left,
              textStyle:
                  GoogleFonts.getFont('Open Sans', fontSize: 14, fontWeight: FontWeight.w400, color: Colors.black)));
    }
    return Container();
  }
}

class DividerParser implements Parser {
  const DividerParser();

  @override
  Widget parse(Map block, {Options options, BuildContext context}) {
    return Divider(
      color: Colors.grey.shade500,
      height: 12,
      thickness: 0.5,
    );
  }
}

class HeaderParser implements Parser {
  const HeaderParser();

  @override
  Widget parse(Map block, {Options options, BuildContext context}) {
    var text = block["text"];
    return PlainTextParser().parse(text,
        context: context,
        options: Options(
            alignment: Alignment.centerLeft,
            padding: EdgeInsets.fromLTRB(0, 8, 0, 4),
            textStyle:
                GoogleFonts.getFont('Open Sans', fontSize: 18, fontWeight: FontWeight.w700, color: Colors.black87)));
  }
}

class ImageParser implements Parser {
  const ImageParser();

  @override
  Widget parse(Map block, {Options options, BuildContext context}) {
    var title = block["title"];
    var url = block["image_url"];
    var innerType = title != null ? title["type"] : "";

    var _image = html.ImageElement(src: url);
    // ignore: undefined_prefixed_name
    ui.platformViewRegistry.registerViewFactory(
      "_image",
      (int viewId) => _image,
    );

    Widget inner = PlainTextParser().parse(title,
        context: context,
        options: Options(
            alignment: Alignment.centerLeft,
            padding: EdgeInsets.fromLTRB(0, 8, 0, 4),
            textStyle: GoogleFonts.getFont('Open Sans',
                fontSize: 14, fontWeight: FontWeight.w400, color: Colors.grey.shade700)));

    return Container(
        child: Column(
      children: [
        inner,
        Container(
            alignment: Alignment.centerLeft,
            child: SizedBox(
                width: _image.width.toDouble(),
                height: _image.height.toDouble(),
                child: HtmlElementView(viewType: "_image"))),
      ],
    ));
  }
}

class DropWidget extends StatefulWidget {
  final void Function(BuildContext context, Map<html.File, Uint8List>) onDropCompleted;
  final Options options;

  DropWidget({this.onDropCompleted, this.options});

  @override
  createState() => DropWidgetState();
}

class DropWidgetState extends State<DropWidget> {
  Map<html.File, Uint8List> files = {};
  bool reading = false;
  bool dragged = false;

  @override
  Widget build(BuildContext context) {
    bool disabled = files.length > 0 || widget.options.disabled;

    return GestureDetector(
        onTap: disabled
            ? null
            : () {
                if (files.length > 0) return;

                html.InputElement uploadInput = html.FileUploadInputElement();
                uploadInput.click();

                uploadInput.onChange.listen((e) {
                  final files = uploadInput.files;
                  onDrop(files);
                  setState(() {
                    dragged = true;
                  });
                });
              },
        child: Container(
            width: 450,
            height: 250,
            padding: EdgeInsets.all(12),
            color: dragged ? Colors.blueGrey.shade100 : Colors.grey.shade200,
            child: Stack(
              children: [
                DottedBorder(
                  color: disabled ? Colors.grey.shade500 : Colors.grey.shade800,
                  dashPattern: dragged ? [8, 0] : [8, 4],
                  strokeCap: StrokeCap.round,
                  strokeWidth: 1,
                  child: DropZone(
                    onDragEnter: disabled
                        ? null
                        : () {
                            setState(() {
                              dragged = true;
                            });
                          },
                    onDragExit: disabled
                        ? null
                        : () {
                            setState(() {
                              dragged = false;
                            });
                          },
                    onDrop: disabled
                        ? null
                        : (List<html.File> files) {
                            onDrop(files);
                          },
                    child: Container(
                      child: Center(
                        child: Column(
                          mainAxisAlignment: MainAxisAlignment.center,
                          children: [
                            Icon(
                              Icons.upload_file,
                              size: 72,
                              color: disabled ? Colors.grey.shade500 : Colors.blueGrey,
                            ),
                            Container(
                                margin: EdgeInsets.only(top: 12, bottom: 12),
                                child: Row(
                                  mainAxisAlignment: MainAxisAlignment.center,
                                  children: getText(),
                                ))
                          ],
                        ),
                      ),
                    ),
                  ),
                ),
                Positioned(
                  right: 4,
                  top: 4,
                  child: SizedBox(
                      width: 32,
                      child: TextButton(
                        onPressed: widget.options.disabled
                            ? null
                            : () {
                                setState(() {
                                  files = {};
                                  reading = false;
                                  dragged = false;
                                });
                              },
                        child: Container(child: Icon(Icons.close)),
                        style: ButtonStyle(
                            foregroundColor:
                                MaterialStateProperty.all(disabled ? Colors.grey.shade500 : Colors.amber.shade700),
                            backgroundColor: MaterialStateProperty.all(Colors.transparent),
                            overlayColor: MaterialStateProperty.all(Colors.transparent),
                            padding: MaterialStateProperty.all(EdgeInsets.zero)),
                      )),
                )
              ],
            )));
  }

  List<Widget> getText() {
    bool disabled = files.length > 0 || widget.options.disabled;

    if (reading) {
      return [
        Text("Preparing files...",
            style: GoogleFonts.getFont('Open Sans', fontSize: 16, fontWeight: FontWeight.w700, color: Colors.blueGrey)),
      ];
    }

    if (files.length == 0) {
      return [
        Text("Choose a file",
            style: GoogleFonts.getFont('Open Sans',
                fontSize: 16, fontWeight: FontWeight.w700, color: disabled ? Colors.grey.shade500 : Colors.blueGrey)),
        Text(" or drag it here.",
            style: GoogleFonts.getFont('Open Sans',
                fontSize: 16, fontWeight: FontWeight.w400, color: disabled ? Colors.grey.shade500 : Colors.blueGrey))
      ];
    }

    List<Widget> names = [
      Text("Done!",
          style: GoogleFonts.getFont('Open Sans', fontSize: 16, fontWeight: FontWeight.w700, color: Colors.blueGrey))
    ];

    for (var file in files.keys) {
      names.add(Text(file.name,
          style: GoogleFonts.getFont('Open Sans', fontSize: 14, fontWeight: FontWeight.w700, color: Colors.blueGrey)));
    }

    return [
      Column(
        children: names,
      )
    ];
  }

  Future<void> onDrop(List<html.File> fileArr) async {
    setState(() {
      reading = true;
    });

    Map<html.File, Uint8List> _files = {};
    for (html.File file in fileArr) {
      html.FileReader reader = html.FileReader();
      reader.readAsArrayBuffer(file);
      await reader.onLoadEnd.first;
      _files[file] = reader.result;
    }

    if (widget.onDropCompleted != null) widget.onDropCompleted(context, _files);

    setState(() {
      reading = false;
      files = _files;
    });
  }
}

class DragDropParser implements Parser {
  Map<html.File, Uint8List> files = {};
  bool reading = false;

  @override
  Widget parse(Map block, {Options options, BuildContext context}) {
    return DropWidget(
      onDropCompleted: this.onDropCompleted,
      options: options ?? Options(),
    );
  }

  void onDropCompleted(BuildContext context, Map<html.File, Uint8List> files) {}
}

class PlainTextInput extends StatefulWidget {
  final Map block;
  final Options options;
  final void Function(BuildContext, String) onSubmitted;

  PlainTextInput(this.block, this.onSubmitted, {this.options});

  @override
  createState() => PlainTextInputState();
}

class PlainTextInputState extends State<PlainTextInput> {
  @override
  Widget build(BuildContext context) {
    var label = widget.block["label"];

    return ConstrainedBox(
        constraints: BoxConstraints(minWidth: 240, maxWidth: 360, minHeight: 64, maxHeight: 64),
        child: Container(
            width: MediaQuery.of(context).size.width * 0.25,
            child: TextField(
              enabled: !widget.options.disabled,
              onSubmitted: (value) {
                widget.onSubmitted(context, value);
              },
              decoration: InputDecoration(
                  border: OutlineInputBorder(), labelText: label, counterText: "\u23CE Press 'enter' to submit"),
            )));
  }
}

class PlainTextInputParser implements Parser {
  @override
  Widget parse(Map block, {Options options, BuildContext context}) {
    return PlainTextInput(
      block,
      this.onSubmitted,
      options: options ?? Options(),
    );
  }

  void onSubmitted(BuildContext context, String value) {}
}

class CheckboxWidget extends StatefulWidget {
  final Map block;
  final Options options;
  final void Function(BuildContext, bool) onChanged;

  CheckboxWidget(this.block, this.onChanged, {this.options});

  @override
  createState() => CheckboxWidgetState();
}

class CheckboxWidgetState extends State<CheckboxWidget> {
  bool value = true;

  @override
  Widget build(BuildContext context) {
    var text = widget.block["text"];
    var description = widget.block["description"];

    return CheckboxListTile(
      value: value,
      controlAffinity: ListTileControlAffinity.leading,
      title: Column(
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          PlainTextParser().parse(text,
              context: context,
              options: Options(
                  textStyle: GoogleFonts.getFont('Open Sans',
                      fontSize: 14, fontWeight: FontWeight.w600, color: Colors.grey.shade800))),
          PlainTextParser().parse(description,
              context: context,
              options: Options(
                  textStyle: GoogleFonts.getFont('Open Sans',
                      fontSize: 14, fontWeight: FontWeight.w500, color: Colors.grey.shade600))),
        ],
      ),
      onChanged: widget.options.disabled
          ? null
          : (_val) {
              setState(() {
                value = _val;
              });
              widget.onChanged(context, _val);
            },
    );
  }
}

class CheckboxParser implements Parser {
  @override
  Widget parse(Map block, {Options options, BuildContext context}) {
    return CheckboxWidget(block, this.onChanged, options: options ?? Options());
  }

  void onChanged(BuildContext context, bool value) {}
}

class RadioButtonWidget extends StatefulWidget {
  final Map block;
  final Options options;
  final void Function(BuildContext, dynamic) onChanged;

  RadioButtonWidget(this.block, this.onChanged, {this.options});

  @override
  createState() => RadioButtonWidgetState();
}

class RadioButtonWidgetState extends State<RadioButtonWidget> {
  String value = "";

  @override
  Widget build(BuildContext context) {
    var options = widget.block["options"];
    List<Widget> radios = [];

    for (var option in options) {
      radios.add(RadioListTile(
          value: option["value"],
          groupValue: value,
          title: PlainTextParser().parse(option["text"],
              context: context,
              options: Options(
                  alignment: Alignment.centerLeft,
                  textStyle: GoogleFonts.getFont('Open Sans', fontSize: 14, fontWeight: FontWeight.w500))),
          onChanged: widget.options.disabled
              ? null
              : (_val) {
                  setState(() {
                    value = _val;
                  });
                  widget.onChanged(context, _val);
                }));
    }

    return Expanded(
      child: Column(
        children: radios,
      ),
    );
  }
}

class RadioButtonParser implements Parser {
  @override
  Widget parse(Map block, {Options options, BuildContext context}) {
    return RadioButtonWidget(block, this.onChanged, options: options ?? Options());
  }

  void onChanged(BuildContext context, value) {}
}

class DateTimeWidget extends StatefulWidget {
  final Map block;
  final Options options;
  final void Function(BuildContext, DateTime) onPicked;

  DateTimeWidget(this.block, this.onPicked, {this.options});

  @override
  createState() => DateTimeWidgetState();
}

class DateTimeWidgetState extends State<DateTimeWidget> {
  DateTime selectedDate = DateTime.now();
  TextEditingController _controller = TextEditingController();
  final _f = new DateFormat("dd.MM.yyyy");

  @override
  void initState() {
    super.initState();
    if (widget.block["initial_date"] != null) selectedDate = DateTime.parse(widget.block["initial_date"]);
    _controller.text = _f.format(selectedDate);
  }

  @override
  Widget build(BuildContext context) {
    return ConstrainedBox(
        constraints: BoxConstraints(minWidth: 240, maxWidth: 240, minHeight: 64, maxHeight: 64),
        child: Container(
            margin: EdgeInsets.all(8),
            width: MediaQuery.of(context).size.width * 0.25,
            child: TextField(
              controller: _controller,
              style: GoogleFonts.getFont('Open Sans',
                  fontSize: 16,
                  fontWeight: FontWeight.w500,
                  color: widget.options.disabled ? Colors.grey.shade500 : Colors.grey.shade800),
              onTap: () async {
                DateTime picked = await showDatePicker(
                    context: context, initialDate: selectedDate, firstDate: DateTime(1900), lastDate: DateTime(2100));

                if (picked != null && picked != selectedDate) {
                  setState(() {
                    selectedDate = picked;
                  });
                  _controller.text = _f.format(selectedDate);
                  widget.onPicked(context, picked);
                }
              },
              enabled: !widget.options.disabled,
              readOnly: true,
              //onSubmitted: widget.onSubmitted,
              decoration: InputDecoration(
                  border: OutlineInputBorder(),
                  labelText: widget.block["placeholder"] != null ? widget.block["placeholder"]["text"] : "",
                  suffixIcon: Container(
                      margin: EdgeInsets.only(right: 6),
                      //decoration: BoxDecoration(color: Colors.grey.shade200),
                      child: Icon(Icons.calendar_today))),
            )));
  }
}

class DateTimePickerParser implements Parser {
  @override
  Widget parse(Map block, {Options options, BuildContext context}) {
    return DateTimeWidget(block, this.onPicked, options: options ?? Options());
  }

  void onPicked(BuildContext context, DateTime date) {}
}
